from __future__ import annotations
import functools, warnings


def deprecated_alias(**aliases):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(func_name, kwargs, aliases):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError(f"{func_name} received both {alias} and {new}")
            warnings.warn(f"{alias} is deprecated; use {new}", DeprecationWarning)
            kwargs[new] = kwargs.pop(alias)


def disabled(func):
    def wrapper(*args, **kwargs):
        raise NotImplementedError(f"The function {func.__name__} is disabled.")

    return wrapper



from logging import getLogger, StreamHandler, INFO

import unittest
import numpy as np

#import openjij as oj
import openjij.cxxjij.graph as G
import openjij.cxxjij.system as S
import openjij.cxxjij.algorithm as A
import openjij.cxxjij.utility as U
import openjij.cxxjij.result as R
from scipy import sparse


class CXXTest(unittest.TestCase):

    def setUp(self):
        self.size = 8
        #dense graph
        self.dense = G.Dense(self.size)
        self.dense = self.gen_testcase(self.dense)
        #sparse graph
        self.sparse = G.Sparse(self.size)
        self.sparse = self.gen_testcase(self.sparse)
        #chimera graph
        #Note: make sure to use ChimeraGPU (not Chimera) when using GPU since the type between FloatType and GPUFloatType is in general different.
        self.chimera = G.ChimeraGPU(2,2)
        self.chimera = self.gen_chimera_testcase(self.chimera)

        self.seed_for_spin = 1234
        self.seed_for_mc = 5678
        

    def gen_testcase(self, J):
        J[0,0]=-0.1
        J[0,1]=-0.9
        J[0,2]=0.2
        J[0,3]=0.1
        J[0,4]=1.3
        J[0,5]=0.8
        J[0,6]=0.9
        J[0,7]=0.4
        J[1,1]=-0.7
        J[1,2]=-1.6
        J[1,3]=1.5
        J[1,4]=1.5
        J[1,5]=1.2
        J[1,6]=-1.5
        J[1,7]=-1.7
        J[2,2]=-0.6
        J[2,3]=1.2
        J[2,4]=-1.3
        J[2,5]=-0.5
        J[2,6]=-1.9
        J[2,7]=1.2
        J[3,3]=0.8
        J[3,4]=-0.5
        J[3,5]=-0.4
        J[3,6]=-1.8
        J[3,7]=-2.0
        J[4,4]=0.6
        J[4,5]=-2.0
        J[4,6]=-1.9
        J[4,7]=0.5
        J[5,5]=-1.8
        J[5,6]=-1.2
        J[5,7]=1.8
        J[6,6]=0.3
        J[6,7]=1.4
        J[7,7]=1.8
        self.true_groundstate = [-1, -1, 1, 1, 1, 1, 1, -1]
        return J

    def gen_testcase_polynomial(self, J):
        J[0, 1, 2, 3, 4] = 0.0686616367121328
        J[0, 2, 3, 4] = 0.0682112165613232
        J[2, 3, 4] = -0.1763027211493039
        J[0, 1, 3, 4] = -0.0907800090462850
        J[1, 3, 4] = 0.1318413458843757
        J[0, 3, 4] = 0.1316587643599703
        J[3, 4] = 0.1460080982070779
        J[4,] = -0.0171180762893237
        J[1, 2, 3] = 0.0137655628870602
        J[0, 2, 4] = 0.1211030013829714
        J[1,] = -0.1487502208910776
        J[0, 1, 2] = 0.0678984161788189
        J[0, 1, 2, 3] = 0.1655848090229992
        J[1, 2, 4] = -0.1628796758769616
        J[3,] = 0.1742156290818721
        J[0, 2, 3] = -0.1081691119002069
        J[1, 4] = 0.1756511179861042
        J[0, 1, 3] = 0.0098192651462946
        J[1, 3] = -0.0746905947645014
        J[0, 3] = 0.1385243673379363
        J[0, 4] = -0.0277205719092218
        J[0, 1, 4] = 0.1113556942155680
        J[0, 2] = -0.0413677095349563
        J[0, 1, 2, 4] = 0.0072610193576964
        J[2,] = -0.1055644094807323
        J[0, 1] = 0.1996162061861095
        J[2, 3] = -0.0226188424784269
        J[1, 2, 3, 4] = 0.0372262067253093
        J[0,] = 0.1730229445472662
        J[2, 4] = 0.0863882044144668
        J[1, 2] = -0.0448357038957756
        J[[]]=0.198873923292106
        self.true_energy = -1.3422641349549371
        return J

    def gen_chimera_testcase(self, J):
        J[0,0,0,G.ChimeraDir.IN_0or4] = +0.25
        J[0,0,0,G.ChimeraDir.IN_1or5] = +0.25
        J[0,0,0,G.ChimeraDir.IN_2or6] = +0.25
        J[0,0,0,G.ChimeraDir.IN_3or7] = +0.25
        J[0,0,1,G.ChimeraDir.IN_0or4] = +0.25
        J[0,0,1,G.ChimeraDir.IN_1or5] = +0.25
        J[0,0,1,G.ChimeraDir.IN_2or6] = +0.25
        J[0,0,1,G.ChimeraDir.IN_3or7] = +0.25
        J[0,0,2,G.ChimeraDir.IN_0or4] = +0.25
        J[0,0,2,G.ChimeraDir.IN_1or5] = +0.25
        J[0,0,2,G.ChimeraDir.IN_2or6] = +0.25
        J[0,0,2,G.ChimeraDir.IN_3or7] = +0.25
        J[0,0,3,G.ChimeraDir.IN_0or4] = +0.25
        J[0,0,3,G.ChimeraDir.IN_1or5] = +0.25
        J[0,0,3,G.ChimeraDir.IN_2or6] = +0.25
        J[0,0,3,G.ChimeraDir.IN_3or7] = +0.25

        J[0,1,0,G.ChimeraDir.IN_0or4] = +0.25
        J[0,1,0,G.ChimeraDir.IN_1or5] = +0.25
        J[0,1,0,G.ChimeraDir.IN_2or6] = +0.25
        J[0,1,0,G.ChimeraDir.IN_3or7] = +0.25
        J[0,1,1,G.ChimeraDir.IN_0or4] = +0.25
        J[0,1,1,G.ChimeraDir.IN_1or5] = +0.25
        J[0,1,1,G.ChimeraDir.IN_2or6] = +0.25
        J[0,1,1,G.ChimeraDir.IN_3or7] = +0.25
        J[0,1,2,G.ChimeraDir.IN_0or4] = +0.25
        J[0,1,2,G.ChimeraDir.IN_1or5] = +0.25
        J[0,1,2,G.ChimeraDir.IN_2or6] = +0.25
        J[0,1,2,G.ChimeraDir.IN_3or7] = +0.25
        J[0,1,3,G.ChimeraDir.IN_0or4] = +0.25
        J[0,1,3,G.ChimeraDir.IN_1or5] = +0.25
        J[0,1,3,G.ChimeraDir.IN_2or6] = +0.25
        J[0,1,3,G.ChimeraDir.IN_3or7] = +0.25

        J[1,0,0,G.ChimeraDir.IN_0or4] = +0.25
        J[1,0,0,G.ChimeraDir.IN_1or5] = +0.25
        J[1,0,0,G.ChimeraDir.IN_2or6] = +0.25
        J[1,0,0,G.ChimeraDir.IN_3or7] = +0.25
        J[1,0,1,G.ChimeraDir.IN_0or4] = +0.25
        J[1,0,1,G.ChimeraDir.IN_1or5] = +0.25
        J[1,0,1,G.ChimeraDir.IN_2or6] = +0.25
        J[1,0,1,G.ChimeraDir.IN_3or7] = +0.25
        J[1,0,2,G.ChimeraDir.IN_0or4] = +0.25
        J[1,0,2,G.ChimeraDir.IN_1or5] = +0.25
        J[1,0,2,G.ChimeraDir.IN_2or6] = +0.25
        J[1,0,2,G.ChimeraDir.IN_3or7] = +0.25
        J[1,0,3,G.ChimeraDir.IN_0or4] = +0.25
        J[1,0,3,G.ChimeraDir.IN_1or5] = +0.25
        J[1,0,3,G.ChimeraDir.IN_2or6] = +0.25
        J[1,0,3,G.ChimeraDir.IN_3or7] = +0.25

        J[1,1,0,G.ChimeraDir.IN_0or4] = +0.25
        J[1,1,0,G.ChimeraDir.IN_1or5] = +0.25
        J[1,1,0,G.ChimeraDir.IN_2or6] = +0.25
        J[1,1,0,G.ChimeraDir.IN_3or7] = +0.25
        J[1,1,1,G.ChimeraDir.IN_0or4] = +0.25
        J[1,1,1,G.ChimeraDir.IN_1or5] = +0.25
        J[1,1,1,G.ChimeraDir.IN_2or6] = +0.25
        J[1,1,1,G.ChimeraDir.IN_3or7] = +0.25
        J[1,1,2,G.ChimeraDir.IN_0or4] = +0.25
        J[1,1,2,G.ChimeraDir.IN_1or5] = +0.25
        J[1,1,2,G.ChimeraDir.IN_2or6] = +0.25
        J[1,1,2,G.ChimeraDir.IN_3or7] = +0.25
        J[1,1,3,G.ChimeraDir.IN_0or4] = +0.25
        J[1,1,3,G.ChimeraDir.IN_1or5] = +0.25
        J[1,1,3,G.ChimeraDir.IN_2or6] = +0.25
        J[1,1,3,G.ChimeraDir.IN_3or7] = +0.2

        J[0,0,0] = +1

        J[0,0,6,G.ChimeraDir.PLUS_C] = +1
        J[0,0,3,G.ChimeraDir.PLUS_R] = -1
        J[1,0,5,G.ChimeraDir.PLUS_C] = +1

        self.true_chimera_spin = [0] * J.size()
        
        self.true_chimera_spin[J.to_ind(0,0,0)] = -1
        self.true_chimera_spin[J.to_ind(0,0,1)] = -1
        self.true_chimera_spin[J.to_ind(0,0,2)] = -1
        self.true_chimera_spin[J.to_ind(0,0,3)] = -1
        self.true_chimera_spin[J.to_ind(0,0,4)] = +1
        self.true_chimera_spin[J.to_ind(0,0,5)] = +1
        self.true_chimera_spin[J.to_ind(0,0,6)] = +1
        self.true_chimera_spin[J.to_ind(0,0,7)] = +1

        self.true_chimera_spin[J.to_ind(0,1,0)] = +1
        self.true_chimera_spin[J.to_ind(0,1,1)] = +1
        self.true_chimera_spin[J.to_ind(0,1,2)] = +1
        self.true_chimera_spin[J.to_ind(0,1,3)] = +1
        self.true_chimera_spin[J.to_ind(0,1,4)] = -1
        self.true_chimera_spin[J.to_ind(0,1,5)] = -1
        self.true_chimera_spin[J.to_ind(0,1,6)] = -1
        self.true_chimera_spin[J.to_ind(0,1,7)] = -1

        self.true_chimera_spin[J.to_ind(1,0,0)] = -1
        self.true_chimera_spin[J.to_ind(1,0,1)] = -1
        self.true_chimera_spin[J.to_ind(1,0,2)] = -1
        self.true_chimera_spin[J.to_ind(1,0,3)] = -1
        self.true_chimera_spin[J.to_ind(1,0,4)] = +1
        self.true_chimera_spin[J.to_ind(1,0,5)] = +1
        self.true_chimera_spin[J.to_ind(1,0,6)] = +1
        self.true_chimera_spin[J.to_ind(1,0,7)] = +1

        self.true_chimera_spin[J.to_ind(1,1,0)] = +1
        self.true_chimera_spin[J.to_ind(1,1,1)] = +1
        self.true_chimera_spin[J.to_ind(1,1,2)] = +1
        self.true_chimera_spin[J.to_ind(1,1,3)] = +1
        self.true_chimera_spin[J.to_ind(1,1,4)] = -1
        self.true_chimera_spin[J.to_ind(1,1,5)] = -1
        self.true_chimera_spin[J.to_ind(1,1,6)] = -1
        self.true_chimera_spin[J.to_ind(1,1,7)] = -1

        return J

    
    def test_SingleSpinFlip_ClassicalIsing_Dense(self):

        #classial ising (dense)
        system = S.make_classical_ising(self.dense.gen_spin(self.seed_for_spin), self.dense)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 100, 100)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    def test_SingleSpinFlip_ClassicalIsing_Sparse(self):

        #classial ising (sparse)
        system = S.make_classical_ising(self.sparse.gen_spin(self.seed_for_spin), self.sparse)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 100, 100)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    def test_SingleSpinFlip_ClassicalIsing_CSRSparse(self):

        #classial ising (csr sparse)
        csr_sparse = G.CSRSparse(sparse.csr_matrix(np.triu(self.dense.get_interactions())))

        system = S.make_classical_ising(self.sparse.gen_spin(self.seed_for_spin), csr_sparse)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 100, 100)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    def test_SingleSpinFlip_Polynomial_Spin(self):
        system_size = 5
        self.polynomial = G.Polynomial(system_size)
        self.polynomial = self.gen_testcase_polynomial(self.polynomial)
        #classial ising (Polynomial)
        system = S.make_classical_ising_polynomial(self.polynomial.gen_spin(), self.polynomial, "SPIN")

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 200, 200)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = system.variables

        #compare
        self.assertAlmostEqual(self.true_energy, self.polynomial.calc_energy(result_spin))

    def test_SingleSpinFlip_Polynomial_Binary(self):
        system_size = 5
        self.polynomial = G.Polynomial(system_size)
        self.polynomial[0]   = +1
        self.polynomial[0,1] = -1
        self.polynomial[0,2] = +1.5
        self.polynomial[0,3] = -1.6
        self.polynomial[0,4] = -1.7
        self.polynomial[1,3] = +2.3
        self.polynomial[1,4] = -0.3
        self.polynomial[2,3] = +3.4
        self.polynomial[2,4] = +3.7
        self.polynomial[3,4] = -0.8
        self.polynomial[0,1,2] = -0.5
        self.polynomial[1,2,3] = -1.0
        self.polynomial[2,3,4] = +0.9
        #classial ising (Polynomial)
        system = S.make_classical_ising_polynomial(self.polynomial.gen_binary(), self.polynomial, "BINARY")

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 200, 200)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = system.variables

        #compare
        self.assertAlmostEqual(-3.1, self.polynomial.calc_energy(result_spin))

    def test_SingleSpinFlip_KLocal_1(self):
        system_size = 5
        self.polynomial = G.Polynomial(system_size)
        self.polynomial[0]   = +1
        self.polynomial[0,1] = -1
        self.polynomial[0,2] = +1.5
        self.polynomial[0,3] = -1.6
        self.polynomial[0,4] = -1.7
        self.polynomial[1,3] = +2.3
        self.polynomial[1,4] = -0.3
        self.polynomial[2,3] = +3.4
        self.polynomial[2,4] = +3.7
        self.polynomial[3,4] = -0.8
        self.polynomial[0,1,2] = -0.5
        self.polynomial[1,2,3] = -1.0
        self.polynomial[2,3,4] = +0.9
        #classial ising (Polynomial)
        system = S.make_k_local_polynomial(self.polynomial.gen_binary(), self.polynomial)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 200, 200)

        #anneal
        A.Algorithm_KLocal_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertAlmostEqual(-3.1, self.polynomial.calc_energy(result_spin))
    
    def test_SingleSpinFlip_KLocal_2(self):
        system_size = 30
        self.polynomial = G.Polynomial(system_size)
        self.polynomial[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29] = -1
        #classial ising (Polynomial)
        system = S.make_k_local_polynomial(self.polynomial.gen_binary(), self.polynomial)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 200, 200)

        #anneal
        A.Algorithm_KLocal_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertAlmostEqual(-1, self.polynomial.calc_energy(result_spin))

    def test_SingleSpinFlip_KLocal_3(self):
        system_size = 30
        self.polynomial = G.Polynomial(system_size)
        self.polynomial[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29] = +1
        self.polynomial[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,   17,18,19,20,21,22,23,24,25,26,27,28,29] = -1
        #classial ising (Polynomial)
        system = S.make_k_local_polynomial(self.polynomial.gen_binary(), self.polynomial)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 200, 200)

        #anneal
        A.Algorithm_KLocal_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertAlmostEqual(-1, self.polynomial.calc_energy(result_spin))

    def test_SingleSpinFlip_KLocal_4(self):
        system_size = 30
        self.polynomial = G.Polynomial(system_size)
        self.polynomial[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29] = -1
        self.polynomial[0,1,2,3,4,5,6,7,8,  10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29] = +1
        self.polynomial[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,   24,25,26,27,28,29] = +1
        self.polynomial[0,1,2,3,4,5,6,7,8,  10,11,12,13,14,15,16,17,18,19,20,21,22,   24,25,26,27,28,29] = -1
        #classial ising (Polynomial)
        system = S.make_k_local_polynomial(self.polynomial.gen_binary(), self.polynomial)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 200, 200)

        #anneal
        A.Algorithm_KLocal_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertAlmostEqual(-1, self.polynomial.calc_energy(result_spin))
    
    def test_SingleSpinFlip_TransverseIsing_Dense(self):

        #transverse ising (dense)
        system = S.make_transverse_ising(self.dense.gen_spin(self.seed_for_spin), self.dense, 1.0, 10)

        #schedulelist
        schedule_list = U.make_transverse_field_schedule_list(10, 100, 100)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    def test_SingleSpinFlip_TransverseIsing_Sparse(self):

        #classial ising (sparse)
        system = S.make_transverse_ising(self.sparse.gen_spin(self.seed_for_spin), self.sparse, 1.0, 10)

        #schedulelist
        schedule_list = U.make_transverse_field_schedule_list(10, 100, 100)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    def test_SingleSpinFlip_TransverseIsing_CSRSparse(self):
        #classial ising (csr sparse)
        csr_sparse = G.CSRSparse(sparse.csr_matrix(np.triu(self.dense.get_interactions())))

        system = S.make_transverse_ising(self.sparse.gen_spin(self.seed_for_spin), csr_sparse, 1.0, 10)

        #schedulelist
        schedule_list = U.make_transverse_field_schedule_list(10, 100, 100)

        #anneal
        A.Algorithm_SingleSpinFlip_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    def test_SwendsenWang_ClassicalIsing_Sparse(self):

        #classial ising (sparse)
        system = S.make_classical_ising(self.sparse.gen_spin(self.seed_for_spin), self.sparse)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 100, 2000)

        #anneal
        A.Algorithm_SwendsenWang_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    def test_SwendsenWang_ClassicalIsing_CSRSparse(self):
        #classial ising (csr sparse)
        csr_sparse = G.CSRSparse(sparse.csr_matrix(np.triu(self.dense.get_interactions())))

        system = S.make_classical_ising(self.sparse.gen_spin(self.seed_for_spin), csr_sparse)

        #schedulelist
        schedule_list = U.make_classical_schedule_list(0.1, 100.0, 100, 2000)

        #anneal
        A.Algorithm_SwendsenWang_run(system, self.seed_for_mc, schedule_list)

        #result spin
        result_spin = R.get_solution(system)

        #compare
        self.assertTrue(self.true_groundstate == result_spin)

    # currently disabled

    #def test_ContinuousTimeSwendsenWang_ContinuousTimeIsing_Sparse(self):

    #    #classial ising (sparse)
    #    system = S.make_continuous_time_ising(self.sparse.gen_spin(self.seed_for_spin), self.sparse, 1.0)

    #    #schedulelist (TODO: why is it so hard?)
    #    schedule_list = U.make_transverse_field_schedule_list(10, 500, 3000)

    #    #anneal
    #    A.Algorithm_ContinuousTimeSwendsenWang_run(system, self.seed_for_mc, schedule_list)

    #    #result spin
    #    result_spin = R.get_solution(system)

    #    #compare
    #    self.assertTrue(self.true_groundstate == result_spin)

# GPU Test is currently disabled.

    # def test_GPU_ChimeraTransverseGPU(self):

    #    #classial ising (sparse)
    #    system = S.make_chimera_transverse_gpu(self.chimera.gen_spin(self.seed_for_spin), self.chimera, 1.0, 10)

    #    #schedulelist
    #    schedule_list = U.make_transverse_field_schedule_list(10, 100, 100)

    #    #anneal
    #    A.Algorithm_GPU_run(system, self.seed_for_mc, schedule_list)

    #    #result spin
    #    result_spin = R.get_solution(system)

    #    #compare
    #    self.assertTrue(self.true_chimera_spin == result_spin)

    # def test_GPU_ChimeraClassicalGPU(self):

    #    #classial ising (sparse)
    #    system = S.make_chimera_classical_gpu(self.chimera.gen_spin(self.seed_for_spin), self.chimera)

    #    #schedulelist
    #    schedule_list = U.make_classical_schedule_list(0.1, 100.0, 100, 100)

    #    #anneal
    #    A.Algorithm_GPU_run(system, self.seed_for_mc, schedule_list)

    #    #result spin
    #    result_spin = R.get_solution(system)

    #    #compare
    #    self.assertTrue(self.true_chimera_spin == result_spin)

#class UtilsTest(unittest.TestCase):
#
#    def test_benchmark(self):
#        h = {0: 1}
#        J = {(0, 1):-1.0, (1,2): -1.0}
#
#        def solver(time_param, iteration):
#            sa_samp = oj.SASampler()
#            sa_samp.step_num = time_param 
#            sa_samp.iteration = iteration
#            return sa_samp.sample_ising(h, J)
#
#        # logger setting
#        logger = getLogger('openjij')
#        stream_handler = StreamHandler()
#        stream_handler.setLevel(INFO)
#        logger.addHandler(stream_handler)
#
#        ground_state = [-1, -1, -1]
#        ground_energy = oj.BinaryQuadraticModel(h, J).calc_energy(ground_state)
#        step_num_list = np.linspace(1, 5, 5, dtype=np.int)
#        bm_res = oj.benchmark([ground_state], ground_energy, solver, time_param_list=step_num_list)
#        self.assertTrue(set(bm_res) >= {'time', 'error', 'e_res', 'tts', 'tts_threshold_prob'})
#
#        self.assertEqual(len(bm_res) ,len(step_num_list))
#
#    def test_response_converter(self):
#        try:
#            from dimod.sampleset import SampleSet
#            import neal
#        except ImportError:
#            print(' skip')
#            return
#        
#        neal_sampler = neal.SimulatedAnnealingSampler()
#        Q = {(1,2):-1, (2,3):-1}
#        response = neal_sampler.sample_qubo(Q)
#        oj_res = oj.convert_response(response)
#
#class CXXTest(unittest.TestCase):
#    def setUp(self):
#        self.N = 10
#        self.dense = cj.graph.Dense(self.N)
#        for i in range(self.N):
#            for j in range(i+1, self.N):
#                self.dense[i, j] = -1
#    def test_cxx_sa(self):
#        sa = cj.system.ClassicalIsing(self.dense)
#        sa.simulated_annealing(beta_min=0.1, beta_max=10.0, step_length=10, step_num=10)
#        ground_spins = sa.get_spins()
#
#        sa.simulated_annealing(schedule=[[0.01, 20]])
#        spins = sa.get_spins()
#
#        self.assertNotEqual(ground_spins, spins)
#
#    def test_cxx_sqa(self):
#        # 1-d model
#        one_d = cj.graph.Dense(self.N)
#        for i in range(self.N):
#            one_d[i, (i+1)%self.N] = -1
#            one_d[i, i] = -1
#        sqa = cj.system.QuantumIsing(one_d, num_trotter_slices=5)
#        sqa.simulated_quantum_annealing(beta=1.0, gamma=2.0, step_length=10, step_num=10)
#        ground_spins = sqa.get_spins()
#
#        sqa.simulated_quantum_annealing(beta=1.0, gamma=2.0, schedule=[[0.5, 200]])
#        spins = sqa.get_spins()
#
#        self.assertNotEqual(ground_spins, spins)
#
#
#
#class ModelTest(unittest.TestCase):
#    def test_bqm(self):
#        h = {}
#        J = {(0,1): -1.0, (1,2): -3.0}
#        bqm = oj.BinaryQuadraticModel(h=h, J=J)
#
#        self.assertEqual(type(bqm.ising_interactions()), np.ndarray)
#        correct_mat = np.array([[0, -1, 0,],[-1, 0, -3],[0, -3, 0]])
#        np.testing.assert_array_equal(bqm.ising_interactions(), correct_mat.astype(np.float))
#
#    def test_chimera_converter(self):
#        h = {}
#        J = {(0,4): -1.0, (6,2): -3.0, (16, 0): 4}
#        chimera = oj.ChimeraModel(h=h, J=J, unit_num_L=2)
#        self.assertEqual(chimera.chimera_coordinate(4, unit_num_L=2), (0,0,4))
#        self.assertEqual(chimera.chimera_coordinate(12, unit_num_L=2), (0,1,4))
#        self.assertEqual(chimera.chimera_coordinate(16, unit_num_L=2), (1,0,0))
#        
#
#    def test_chimera(self):
#        h = {}
#        J = {(0,4): -1.0, (6,2): -3.0}
#        bqm = oj.ChimeraModel(h=h, J=J, unit_num_L=3)
#        self.assertTrue(bqm.validate_chimera())
#
#        J = {(0, 1): -1}
#        bqm = oj.ChimeraModel(h=h, J=J, unit_num_L=3)
#        with self.assertRaises(ValueError):
#            bqm.validate_chimera()
#        
#        J = {(4, 12): -1}
#        bqm = oj.ChimeraModel(h=h, J=J, unit_num_L=2)
#        self.assertTrue(bqm.validate_chimera())
#
#        J = {(0,4): -1, (5, 13):1, (24, 8):2, (18,20): 1, (16,0):0.5, (19, 23): -2}
#        h = {13: 2}
#        chimera = oj.ChimeraModel(h, J, unit_num_L=2)
#        self.assertEqual(chimera.to_index(1,1,1, unit_num_L=2), 25)
#
#        self.assertTrue(chimera.validate_chimera())
#
#
#
#    def test_ising_dict(self):
#        Q = {(0,4): -1.0, (6,2): -3.0}
#        bqm = oj.ChimeraModel(Q=Q, vartype='BINARY', unit_num_L=3)
#
#    def test_king_graph(self):
#        h = {}
#        J = {(0,1): -1.0, (1,2): -3.0}
#        king_interaction = [[0,0, 1,0, -1.0], [1,0, 2,0, -3.0]]
#
#        king_graph = oj.KingGraph(machine_type="ASIC", h=h, J=J)
#        correct_mat = np.array([[0, -1, 0,],[-1, 0, -3],[0, -3, 0]])
#        np.testing.assert_array_equal(king_graph.ising_interactions(), correct_mat.astype(np.float))
#        np.testing.assert_array_equal(king_interaction, king_graph._ising_king_graph)
#
#        king_graph = oj.KingGraph(machine_type="ASIC", king_graph=king_interaction)
#        np.testing.assert_array_equal(king_interaction, king_graph._ising_king_graph)
#
#
#        king_graph = oj.KingGraph(machine_type="ASIC", Q={(0,1): -1}, vartype="BINARY")
#        king_interaction = [[0, 0, 0, 0, -0.25], [0,0,1,0,-0.25], [1,0,1,0,-0.25]]
#        np.testing.assert_array_equal(king_interaction, king_graph._ising_king_graph)
#
#class TestChimeraGraph(unittest.TestCase):
#    def full_chimera_qubo(self, L):
#
#        left_side = [0,1,2,3]
#        right_side = [4,5,6,7]
#        to_ind = lambda r,c,i: 8*L*r + 8*c + i
#        Q = {}
#        # Set to -1 for all bonds in each chimera unit
#        for c in range(L):
#            for r in range(L):
#                for z_l in left_side:
#                    for z_r in right_side:
#                        Q[to_ind(r,c,z_l), to_ind(r,c,z_r)] = -1
#
#                        # linear term
#                        Q[to_ind(r,c,z_l), to_ind(r,c,z_l)] = -1
#                    #linear term
#                    Q[to_ind(r,c,z_r), to_ind(r,c,z_r)] = -1
#
#        # connect all chimera unit
#        # column direction
#        for c in range(L-1):
#            for r in range(L):
#                for z_r in right_side:
#                    Q[to_ind(r,c,z_r), to_ind(r,c+1,z_r)] = +0.49
#        # row direction
#        for r in range(L-1):
#            for c in range(L):
#                for z_l in left_side:
#                    Q[to_ind(r,c,z_l), to_ind(r+1,c,z_l)] = 0.49
#        return Q
#
#    def full_chimera_ising(self, L):
#        Q = self.full_chimera_qubo(L)
#        h, J = {}, {}
#        for (i, j), value in Q.items():
#            if i == j:
#                h[i] = value
#            else:
#                J[i, j] = value
#        return h, J
#
#    def test_chimera_validate(self):
#        L = 4
#        Q = self.full_chimera_qubo(L=L)
#        chimera = oj.ChimeraModel(Q=Q, unit_num_L=L, vartype='BINARY')
#
#        self.assertTrue(chimera._validate((0,0,0),(0,0,4),L))
#        self.assertFalse(chimera._validate((0,0,0),(96,0,0),L))
#
#
#
#    def test_chimera_connect(self):
#        Q = self.full_chimera_qubo(L=2)
#        chimera = oj.ChimeraModel(Q=Q, unit_num_L=2, vartype='BINARY')
#        self.assertTrue(chimera.validate_chimera())
#
#        Q = self.full_chimera_qubo(L=4)
#        chimera = oj.ChimeraModel(Q=Q, unit_num_L=4, vartype='BINARY')
#        self.assertTrue(chimera.validate_chimera())
#
#
if __name__ == '__main__':
    unittest.main()






from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

from openjij.sampler.chimera_gpu.gpu_sa_sampler import GPUChimeraSASampler
from openjij.sampler.chimera_gpu.gpu_sqa_sampler import GPUChimeraSQASampler

__all__ = ["GPUChimeraSASampler", "GPUChimeraSQASampler"]



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import openjij.model.chimera_model, openjij.model.king_graph

from openjij.model.model import (
    BinaryPolynomialModel,
    BinaryQuadraticModel,
    bqm_from_numpy_matrix,
    make_BinaryPolynomialModel,
    make_BinaryPolynomialModel_from_hising,
    make_BinaryPolynomialModel_from_hubo,
    make_BinaryPolynomialModel_from_JSON,
    make_BinaryQuadraticModel,
    make_BinaryQuadraticModel_from_JSON,
)

__all__ = [
    "make_BinaryQuadraticModel",
    "make_BinaryQuadraticModel",
    "BinaryQuadraticModel",
    "bqm_from_numpy_matrix",
    "make_BinaryPolynomialModel",
    "make_BinaryPolynomialModel_from_JSON",
    "BinaryPolynomialModel",
    "make_BinaryPolynomialModel_from_hising",
    "make_BinaryPolynomialModel_from_hubo",
]



from __future__ import annotations
import numpy as np

import openjij
import openjij as oj
import openjij.cxxjij as cxxjij

from openjij.sampler.sqa_sampler import SQASampler


class CSQASampler(SQASampler):
    """Sampler with continuous-time simulated quantum annealing (CSQA) using

    Hamiltonian.

    .. math::

        H(s) = s H_p + \\Gamma (1-s)\\sum_i \\sigma_i^x

    where :math:`H_p` is the problem Hamiltonian we want to solve.

    Args:
        beta (float): Inverse temperature.
        gamma (float): Amplitude of quantum fluctuation.
        schedule (list): schedule list
        step_num (int): Number of Monte Carlo step.
        schedule_info (dict): Information about a annealing schedule.
        num_reads (int): Number of iterations.
        num_sweeps (int): number of sweeps
        schedule_info (dict): Information about a annealing schedule.
    """

    def __init__(
        self, beta=5.0, gamma=1.0, num_sweeps=1000, schedule=None, num_reads=1
    ):

        self._default_params = {
            "beta": beta,
            "gamma": gamma,
            "num_sweeps": num_sweeps,
            "schedule": schedule,
            "num_reads": num_reads,
        }

        self._params = {
            "beta": beta,
            "gamma": gamma,
            "num_sweeps": num_sweeps,
            "schedule": schedule,
            "num_reads": num_reads,
        }

    def _get_result(self, system, model):
        info = {}
        info["spin_config"] = system.spin_config

        state = cxxjij.result.get_solution(system)

        return state, info

    def sample_ising(
        self,
        h,
        J,
        beta=None,
        gamma=None,
        num_sweeps=None,
        schedule=None,
        num_reads=None,
        initial_state=None,
        updater=None,
        reinitialize_state=True,
        seed=None,
    ):
        """Sampling from the Ising model.

        Args:
            h (dict): linear biases
            J (dict): quadratic biases
            beta (float, optional): inverse temperature
            gamma (float, optional): strength of transverse field
            num_sweeps (int, optional): number of sampling.
            schedule (list, optional): schedule list
            num_reads (int, optional): number of iterations
            initial_state (optional): initial state of spins
            updater (str, optional): updater algorithm
            reinitialize_state (bool, optional): Re-initilization at each sampling. Defaults to True.
            seed (int, optional): Sampling seed.

        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples:

            for Ising case::

                >>> h = {0: -1, 1: -1, 2: 1, 3: 1}
                >>> J = {(0, 1): -1, (3, 4): -1}
                >>> sampler = openjij.CSQASampler()
                >>> res = sampler.sample_ising(h, J)

            for QUBO case::

                >>> Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1, (0, 1): -1, (3, 4): 1}
                >>> sampler = openjijj.CSQASampler()
                >>> res = sampler.sample_qubo(Q)
        """

        # Set default updater
        if updater is None:
            updater = "swendsenwang"

        bqm = oj.model.model.BinaryQuadraticModel(
            linear=h, quadratic=J, vartype="SPIN", sparse=True
        )

        # Continuous time ising system only supports sparse ising graph

        ising_graph = bqm.get_cxxjij_ising_graph()

        self._set_params(
            beta=beta, gamma=gamma, num_sweeps=num_sweeps, num_reads=num_reads
        )
        self._annealing_schedule_setting(
            bqm,
            self._params["beta"],
            self._params["gamma"],
            self._params["num_sweeps"],
            self._params["schedule"],
        )

        # make init state generator --------------------------------
        if initial_state is None:

            def init_generator():
                spin_config = np.random.choice([1, -1], len(bqm.variables))
                return list(spin_config)

        else:

            def init_generator():
                return initial_state

        # -------------------------------- make init state generator

        # choose updater -------------------------------------------
        sqa_system = cxxjij.system.make_continuous_time_ising(
            init_generator(), ising_graph, self.gamma
        )
        _updater_name = updater.lower().replace("_", "").replace(" ", "")
        if _updater_name == "swendsenwang":
            algorithm = cxxjij.algorithm.Algorithm_ContinuousTimeSwendsenWang_run
        else:
            raise ValueError('updater is one of "swendsen wang"')
        # ------------------------------------------- choose updater

        response = self._cxxjij_sampling(
            bqm, init_generator, algorithm, sqa_system, reinitialize_state, seed
        )

        response.info["schedule"] = self.schedule_info

        return response



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import dimod

SPIN = dimod.SPIN
BINARY = dimod.BINARY

Vartype = dimod.Vartype


def cast_vartype(vartype):
    if isinstance(vartype, dimod.Vartype):
        return vartype
    elif isinstance(vartype, str):
        if vartype.upper() == "SPIN":
            return SPIN
        elif vartype.upper() == "BINARY":
            return BINARY



# Copyright 2021 Jij Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

from __future__ import annotations
import dimod

from openjij.utils.graph_utils import chimera_to_ind


class BaseGPUChimeraSampler(dimod.Structured):
    """Abstract GPUChimera Sampler."""

    @property
    def parameters(self):
        param = super().parameters
        param["unit_num_L"] = ["system parameter"]
        return param

    @property
    def nodelist(self):
        if "unit_num_L" not in dir(self) or self.unit_num_L is None:
            raise ValueError("this object is not set unit_num_L yet.")

        return list(range(0, 8 * 8 * self.unit_num_L * self.unit_num_L))

    @property
    def edgelist(self):
        if "unit_num_L" not in dir(self) or self.unit_num_L is None:
            raise ValueError("this object is not set unit_num_L yet.")

        edges = []
        for r in range(self.unit_num_L):
            for c in range(self.unit_num_L):
                for z in range(8):
                    i = chimera_to_ind(r, c, z, self.unit_num_L)

                    if z in [0, 1, 2, 3]:
                        edges += [
                            (i, chimera_to_ind(r, c, _z, self.unit_num_L))
                            for _z in [4, 5, 6, 7]
                        ]
                        if r < self.unit_num_L - 1:
                            edges.append(
                                (i, chimera_to_ind(r + 1, c, z, self.unit_num_L))
                            )

                    elif z in [4, 5, 6, 7]:
                        if c < self.unit_num_L - 1:
                            edges.append(
                                (i, chimera_to_ind(r, c + 1, z, self.unit_num_L))
                            )
        return edges



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

from openjij.sampler.csqa_sampler import CSQASampler
from openjij.sampler.response import Response
from openjij.sampler.sa_sampler import SASampler
from openjij.sampler.sampler import measure_time
from openjij.sampler.sqa_sampler import SQASampler

__all__ = [
    "CSQASampler",
    "Response",
    "SASampler",
    "measure_time",
    "SQASampler",
]



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from dimod import SPIN

import openjij
import openjij as oj

from openjij.model.model import make_BinaryQuadraticModel


def make_KingGraph(linear=None, quadratic=None, king_graph=None):
    """KingGraph factory

    Returns:
        generated KingGraph class
    """
    mock_linear = {}
    mock_quadratic = {}

    if linear is not None:
        mock_linear = linear

    if quadratic is not None:
        mock_quadratic = quadratic

    if mock_linear == {} and mock_quadratic == {}:
        # no elements in linear and quadratic
        # fetch first element of the king_graph
        label = (king_graph[0][0], king_graph[0][0])
        # add to linear
        mock_linear[label] = 1.0

    class KingGraph(make_BinaryQuadraticModel(mock_linear, mock_quadratic)):
        """
        BQM for king graph of HITACHI CMOS Annealer

        Attributes:
            xrange (list(int)): represents hardware (CMOS) restricts for coordinate. [xmin, xmax]
            yrange (list(int)): represents hardware (CMOS) restricts for coordinate. [ymin, ymax]
            prange (list(int)): represents hardware (CMOS) restricts for the strength of interactions 'p'. [pmin, pmax]
            king_graph (list(int)):
                Annealing cloud Web API format representation of interaction coefficients
                Quadratic term [x1, y1, x2, y2, value]
                Linear term    [x1, y1, x1, y1, value]
        """

        def __init__(
            self,
            linear: dict = None,
            quadratic: dict = None,
            offset: float = 0.0,
            king_graph=None,
            vartype=SPIN,
            machine_type: str = "",
        ):
            """__init__.

            Args:
                linear (dict): linear biases
                quadratic (dict): quadratic biases
                offset (float): offset
                king_graph: represents ising or QUBO interaction.
                    Each spins are decided by 2-d corrdinate x, y.
                    Quadratic term [x1, y1, x2, y2, value]
                    Linear term    [x1, y1, x1, y1, value]
                vartype: 'SPIN' or 'BINARY'
                machine_type (str): choose 'ASIC' or 'FPGA'
            """
            vartype = oj.variable_type.cast_vartype(vartype)

            # set parameter ranges
            self.machine_type = machine_type
            if self.machine_type == "ASIC":
                self.xrange = [0, 351 + 1]
                self.yrange = [0, 175 + 1]
                self.prange = [-3, 3]
            elif self.machine_type == "FPGA":
                self.xrange = [0, 79 + 1]
                self.yrange = [0, 79 + 1]
                self.prange = [-127, 127]
            else:
                raise ValueError("machine type should be ASIC or FPGA")

            # convert format h, J, Q and initilize BQM
            if king_graph is not None:
                linear, quadratic = self._convert_to_BQM_format(king_graph, vartype)
            super().__init__(linear, quadratic, offset=offset, vartype=vartype)

            # reformat to ising king graph (which is Web API format)
            if king_graph is not None and vartype == SPIN:
                self._ising_king_graph = king_graph

            else:
                # generate Ising h and J and create ising_king_graph format
                lin, quad, _ = self.to_ising()
                self._ising_king_graph = []
                for index, h in lin.items():
                    if h != 0:
                        x, y = self._convert_to_xy(index)
                        self._ising_king_graph.append([x, y, x, y, h])
                for (i, j), J in quad.items():
                    if J != 0:
                        x1, y1 = self._convert_to_xy(i)
                        x2, y2 = self._convert_to_xy(j)
                        self._ising_king_graph.append([x1, y1, x2, y2, J])

            self._validation_ising_king_graph()

        def _convert_to_BQM_format(self, king_graph, vartype):
            linear, quad = {}, {}
            for x1, y1, x2, y2, value in king_graph:
                if (x1, y1) == (x2, y2):
                    linear[(x1, y1)] = value
                else:
                    quad[(x1, y1), (x2, y2)] = value
            return linear, quad

        def get_ising_king_graph(self):
            return self._ising_king_graph

        def king_indices(self):
            if isinstance(self.indices[0], tuple):
                return self.indices
            else:
                return [self._convert_to_xy(i) for i in self.indices]

        def _convert_to_xy(self, index):
            if isinstance(index, tuple):
                return index[0], index[1]
            else:
                y = int(index / self.xrange[1])
                return int(index - y * self.xrange[1]), y

        def convert_to_index(self, x, y):
            return y * self.xrange[1] + x

        def _validation_ising_king_graph(self):
            for xi, yi, xj, yj, p in self._ising_king_graph:
                if yi >= self.yrange[1] or yj >= self.yrange[1]:
                    raise ValueError(
                        "Graph is incomplete xi: {}, yi: {}, xj: {}, yj: {}, p:{}".format(
                            xi, yi, xj, yj, p
                        )
                    )
                if not (xi in [xj, xj - 1, xj + 1]) or not (yi in [yj, yj - 1, yj + 1]):
                    raise ValueError(
                        "Graph is incomplete xi: {}, yi: {}, xj: {}, yj: {}, p:{}".format(
                            xi, yi, xj, yj, p
                        )
                    )
                if not (self.prange[0] <= p <= self.prange[1]):
                    raise ValueError(
                        "Graph is incomplete xi: {}, yi: {}, xj: {}, yj: {}, p: {}".format(
                            xi, yi, xj, yj, p
                        )
                    )

        def energy(self, sample):
            return super().energy(sample, sparse=True)

        def energies(self, samples_like):
            return super().energies(sample_like, sparse=True)

    return KingGraph


def make_KingGraph_from_JSON(obj):
    """KingGraph factory for JSON

    Args:
        obj (dict): JSON object
    Returns:
        generated KingGraph class
    """
    label = obj["variable_labels"][0]
    if isinstance(label, list):
        # convert to tuple
        label = tuple(label)

    mock_linear = {label: 1.0}

    return make_KingGraph(mock_linear, {})


def KingGraph(
    linear=None,
    quadratic=None,
    offset=0.0,
    king_graph=None,
    vartype=SPIN,
    machine_type="",
):
    """Generate KingGraph model.

    Args:
        linear (dict): linear biases
        quadratic (dict): quadratic biases
        offset (float): offset
        king_graph: represents ising or QUBO interaction.
            Each spins are decided by 2-d corrdinate x, y.

            * Quadratic term: [x1, y1, x2, y2, value]
            * Linear term:    [x1, y1, x1, y1, value]
        vartype: 'SPIN' or 'BINARY'
        machine_type (str): choose 'ASIC' or 'FPGA'
    Returns:
        generated KingGraphModel
    Examples:
        The following code shows intialization of KingGraph::

            >>> h = {}
            >>> J = {(0, 1): -1.0, (1, 2): -3.0}
            >>> king_graph = oj.KingGraph(machine_type="ASIC", linear=h, quadratic=J)

        You can initialize it from `king_interaction`::

            >>> king_interaction = [[0, 0, 1, 0, -1.0], [1, 0, 2, 0, -3.0]]
            >>> king_graph = oj.KingGraph(machine_type="ASIC", king_graph=king_interaction)
    """

    Model = make_KingGraph(linear, quadratic, king_graph)

    return Model(linear, quadratic, offset, king_graph, vartype, machine_type)


# classmethods
KingGraph.from_qubo = lambda Q, offset=0.0, **kwargs: make_KingGraph({}, Q).from_qubo(
    Q, offset, **kwargs
)

KingGraph.from_ising = lambda linear, quadratic, offset=0.0, **kwargs: make_KingGraph(
    linear, quadratic
).from_ising(linear, quadratic, offset, **kwargs)

KingGraph.from_serializable = lambda obj: make_KingGraph_from_JSON(
    obj
).from_serializable(obj)



# Copyright 2022 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

try:
    from skbuild import setup
except ImportError:
    from setuptools import setup

setup_requires = [
    "numpy",
    "pybind11 >=2.10.0, < 2.11.0",
    "cmake > 3.20",
    "scikit-build > 0.16.0"
]

if any(arg in sys.argv for arg in ("pytest", "test")):
    setup_requires.append("pytest-runner")

setup(
    setup_requires=setup_requires,
    packages=[
        "openjij",
        "openjij.model",
        "openjij.sampler",
        "openjij.sampler.chimera_gpu",
        "openjij.utils",
    ],
    cmake_install_dir="openjij",
    include_package_data=False,
    zip_safe=False,
)



import unittest
import openjij as oj


class TestSQA(unittest.TestCase):
    def setUp(self):
        self.num_ind = {
            'h': {0: -1, 1: -1, 2: 1, 3: 1},
            'J': {(0, 1): -1, (3, 4): -1}
        }
        str_ind = ['a', 'b', 'c', 'd', 'e']
        self.str_ising = {
            'h': {str_ind[i] for i in self.num_ind['h'].keys()},
            'J': {(str_ind[i], str_ind[j]) for i, j in self.num_ind['J'].keys()}
        }
        self.ground_state = [1, 1, -1, -1, -1]
        self.e_g = -1-1-1-1 + (-1-1)
        self.g_sample = {i: self.ground_state[i]
                         for i in range(len(self.ground_state))}
        self.g_samp_str = {k: self.ground_state[i]
                           for i, k in enumerate(str_ind)}

        self.qubo = {
            (0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1,
            (0, 1): -1, (3, 4): 1
        }
        self.str_qubo = {(str_ind[i], str_ind[j]): qij
                         for (i, j), qij in self.qubo.items()}
        self.ground_q = [1, 1, 0, 0, 0]
        self.e_q = -1-1-1

    #def test_sqa(self):
    #    sampler = oj.SQASampler()



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import openjij as oj


def convert_response(response):
    if isinstance(response, oj.sampler.response.Response):
        return response

    try:
        from dimod.response import Response
        from dimod.sampleset import SampleSet
    except ImportError:
        raise ImportError('Import dwave dimod : "pip install dimod"')

    if isinstance(response, Response) or isinstance(response, SampleSet):
        from dimod.vartypes import BINARY

        vartype = "BINARY" if response.vartype == BINARY else "SPIN"
        o_res = oj.sampler.response.Response(
            vartype=vartype, indices=list(response.variables)
        )
        states = []
        energies = []
        for rec in response.record:
            for _ in range(rec[2]):
                states.append(rec[0])
                energies.append(rec[1])
        o_res.update_ising_states_energies(states=states, energies=energies)

        o_res.info = response.info
        if "qpu_sampling_time" in response.info:
            o_res.info["sampling_time"] = response.info["qpu_sampling_time"]
        if "anneal_time_per_run" in response.info:
            o_res.info["execution_time"] = response.info["anneal_time_per_run"]
        o_res.info["dimod"] = response
        return o_res
    else:
        raise TypeError(
            "response is dimod.response.Response, SampleSet or openjij.sampler.response.Response"
        )
